# -*- coding: utf-8 -*-
import yagmail


class UtilMail(object):
    def __init__(self, user, password, host, port):
        self.__yag = yagmail.SMTP(user=user, password=password, host=host, port=port)

    def send(self, to, subject, contents: list, cc=None, attachments=None):
        """发送"""
        self.__yag.send(to=to, subject=subject, contents=contents, cc=cc, attachments=attachments)



# -*- coding: utf-8 -*-
import os
import sys
import time
import string
import zipfile
import logging
import undetected_chromedriver as uc

from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.select import Select
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver import Remote
from selenium.common.exceptions import InvalidArgumentException
from selenium.webdriver.chrome import options

from wise_utils.spider.basespider import BaseSpider


class ReuseChrome(Remote):
    """
    用于重连已存在的浏览器，由于使用官方方法重连时，会先建一个新的浏览器，再修改 executor_url 和 session_id，多创建一个空白浏览器窗口，因此。继承并重写其start_session方法，
    """

    def __init__(self, command_executor, session_id):
        self.r_session_id = session_id
        Remote.__init__(self, command_executor=command_executor, desired_capabilities={})

    def start_session(self, capabilities, browser_profile=None):
        """
        重写start_session方法
        """
        if not isinstance(capabilities, dict):
            raise InvalidArgumentException("Capabilities must be a dictionary")
        if browser_profile:
            if "moz:firefoxOptions" in capabilities:
                capabilities["moz:firefoxOptions"]["profile"] = browser_profile.encoded
            else:
                capabilities.update({'firefox_profile': browser_profile.encoded})

        self.caps = options.Options().to_capabilities()
        self.session_id = self.r_session_id
        self.w3c = False


class SeleniumBaseSpider(BaseSpider):
    PAGE_LOAD_TIMEOUT = 180

    def __init__(self, task_name, log_level=logging.DEBUG, log_file_path=None):
        super(SeleniumBaseSpider, self).__init__(task_name=task_name, log_level=log_level, log_file_path=log_file_path)
        self.__browser_model = "local"  # 默认为本地启动

    def create_local_browser(self,
                             headless: bool = True,
                             incognito: bool = True,
                             no_picture: bool = False,
                             no_pop_ups: bool = False,
                             page_load_strategy: str = "none",
                             no_extensions: bool = False,
                             use_proxy: bool = False,
                             userdata_file_path: str = "",
                             executable_path: str = "",
                             chrome_options: webdriver.ChromeOptions = None,
                             execute_cdp_cmd: bool = False
                             ):
        """
        创建本地浏览器
        :params headless: 无头模式
        :params incognito: 无痕模式
        :params no_picture: 禁止加载图片元素
        :params no_pop_ups: 禁止弹窗
        :params page_load_strategy: 页面加载策略 可设置为三种  1 eager selenium会等待整个界面加载完成（指对html和子资源的下载与解析，不包括ajax）  2 normal 要等待整个dom树加载完成，即DOMContentLoaded这个事件完成，仅对html的内容进行下载解析 3 none 当html下载完成之后，不等待解析完成，selenium会直接返回
        :params no_extensions: 禁止使用插件
        :params use_proxy: 使用代理
        :params userdata_file_path: 保存历史和缓存（多爬虫同时使用不行）
        """
        self.__browser_model = "local"
        if sys.platform == "linux":
            # 打开虚拟窗口
            from pyvirtualdisplay import Display
            self.__display = Display(visible=False, size=(1792, 1120))
            self.__display.start()

        chrome_options = chrome_options if chrome_options else self.get_chrome_options(
            headless=headless,
            incognito=incognito,
            no_picture=no_picture,
            no_pop_ups=no_pop_ups,
            page_load_strategy=page_load_strategy,
            no_extensions=no_extensions,
            use_proxy=use_proxy,
            userdata_file_path=userdata_file_path
        )

        if executable_path:
            service = Service(executable_path=executable_path)
        else:
            # 自动下载
            from webdriver_manager.chrome import ChromeDriverManager
            service = Service(executable_path=ChromeDriverManager().install())

        browser = webdriver.Chrome(service=service, options=chrome_options)

        if execute_cdp_cmd:
            # 隐藏浏览器特征
            with open(os.path.join(os.path.dirname(__file__), "stealth.js")) as f:
                js = f.read()
            browser.execute_cdp_cmd("Page.addScriptToEvaluateOnNewDocument", {"source": js})

        browser.set_page_load_timeout(self.PAGE_LOAD_TIMEOUT)
        browser.maximize_window()
        return browser

    def create_remote_browser(self, remote_url,
                              desired_capabilities: dict = {},
                              headless: bool = True,
                              incognito: bool = True,
                              no_picture: bool = False,
                              no_pop_ups: bool = False,
                              page_load_strategy: str = "none",
                              no_extensions: bool = False,
                              use_proxy: bool = False,
                              userdata_file_path: str = "",
                              chrome_options: webdriver.ChromeOptions = None,
                              ):
        """
        创建远程浏览器: selenium grid
        :params headless: 无头模式
        :params incognito: 无痕模式
        :params no_picture: 禁止加载图片元素
        :params no_pop_ups: 禁止弹窗
        :params page_load_strategy: 页面加载策略 可设置为三种  1 eager selenium会等待整个界面加载完成（指对html和子资源的下载与解析，不包括ajax）  2 normal 要等待整个dom树加载完成，即DOMContentLoaded这个事件完成，仅对html的内容进行下载解析 3 none 当html下载完成之后，不等待解析完成，selenium会直接返回
        :params no_extensions: 禁止使用插件
        :params use_proxy: 使用代理
        :params userdata_file_path: 保存历史和缓存（多爬虫同时使用不行）
        """

        self.__browser_model = "remote"
        chrome_options = chrome_options if chrome_options else self.get_chrome_options(
            headless=headless,
            incognito=incognito,
            no_picture=no_picture,
            no_pop_ups=no_pop_ups,
            page_load_strategy=page_load_strategy,
            no_extensions=no_extensions,
            use_proxy=use_proxy,
            userdata_file_path=userdata_file_path
        )

        desired_capabilities = desired_capabilities if desired_capabilities else {
            "browserName": "chrome",
            "platformName": "linux",
            "se:recordVideo": "true",
            "se:timeZone": "Asia/Shanghai",
            "se:screenResolution": "1920x1080"
        }
        browser = webdriver.Remote(command_executor=remote_url, desired_capabilities=desired_capabilities, options=chrome_options)
        browser.set_page_load_timeout(self.PAGE_LOAD_TIMEOUT)
        browser.maximize_window()
        return browser

    def create_undetected_browser(self,
                                  chrome_version,
                                  headless: bool = True,
                                  incognito: bool = True,
                                  no_pop_ups: bool = False,
                                  page_load_strategy: str = "none",
                                  no_extensions: bool = False,
                                  userdata_file_path: str = "",
                                  chrome_options: uc.ChromeOptions = None,

                                  ):
        """通过undetected_chromedriver创建chrome，需要指定浏览器版本。可过五秒盾"""
        self.__browser_model = "undetected"

        if sys.platform == "linux":
            # 打开虚拟窗口
            from pyvirtualdisplay import Display
            self.__display = Display(visible=False, size=(1792, 1120))
            self.__display.start()

        from multiprocessing import freeze_support
        freeze_support()

        chrome_options = chrome_options if chrome_options else self.get_undetected_chrome_options(
            headless=headless,
            incognito=incognito,
            no_pop_ups=no_pop_ups,
            page_load_strategy=page_load_strategy,
            no_extensions=no_extensions,
            userdata_file_path=userdata_file_path
        )
        browser = uc.Chrome(options=chrome_options, version_main=chrome_version)

        browser.set_page_load_timeout(self.PAGE_LOAD_TIMEOUT)
        browser.maximize_window()
        return browser

    @staticmethod
    def reconnect_existed_browser(executor_url, session_id):
        """
        重新连接一个已经存在的浏览器
        :params executor_url: browser.command_executor._url
        :params session_id: browser.session_id
        """
        return ReuseChrome(command_executor=executor_url, session_id=session_id)

    def get_chrome_options(self,
                           headless: bool = True,
                           incognito: bool = True,
                           no_picture: bool = False,
                           no_pop_ups: bool = False,
                           page_load_strategy: str = "none",
                           no_extensions: bool = False,
                           use_proxy: bool = False,
                           userdata_file_path: str = "",
                           ):
        """获取chrome配置信息"""
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument("lang=zh_CN.UTF-8")  # 设置中文
        # chrome_options.add_argument(f"user-agent={self.get_random_ua()}")  # 更换头部
        chrome_options.add_argument("--no-sandbox")  # root
        chrome_options.add_argument("--disable-gpu")  # 谷歌文档提到需要加上这个属性来规避bug
        chrome_options.add_argument("--hide-scrollbars")  # 隐藏滚动条, 应对一些特殊页面
        chrome_options.add_argument("--disable-infobars")  # 关闭提示
        chrome_options.add_argument("--disable-dev-shm-usage")  # 修复选项卡崩溃
        chrome_options.add_argument("--ignore-certificate-errors")  # 忽略验证：您的连接不是私密连接
        chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])  # 取消chrome受自动控制提示
        # chrome_options.add_argument("–single-process")  # 单进程运行Google Chrome
        chrome_options.add_experimental_option('useAutomationExtension', False)
        chrome_options.add_argument("--disable-blink-features=AutomationControlled")

        if no_extensions:
            chrome_options.add_argument("--disable-extensions")
            chrome_options.add_experimental_option("useAutomationExtension", False)

        if no_picture:
            chrome_options.add_argument("blink-settings=imagesEnabled=false")

        if no_pop_ups:
            # 禁止弹窗加入
            prefs = {
                "profile.default_content_setting_values":
                    {
                        "notifications": 2
                    }
            }
        else:
            prefs = dict()
        prefs["credentials_enable_service"] = False
        prefs["profile.password_manager_enabled"] = False
        chrome_options.add_experimental_option("prefs", prefs)
        chrome_options.set_capability("PageloadStrategy", page_load_strategy)

        if use_proxy:
            # 无头模式，无痕模式 不支持使用扩展
            headless = False
            incognito = False
            use_userdata = False
            # 获取代理
            proxy_dict = self.get_proxy_ip()
            if proxy_dict.get("username", None) is not None:
                # 创建插件
                plugin_path = self.__create_proxyauth_extension(proxy_host=proxy_dict["ip"], proxy_port=proxy_dict["port"], proxy_username=proxy_dict["username"], proxy_password=proxy_dict["password"])
                chrome_options.add_extension(plugin_path)
            else:
                # 代理不需要账号密码
                chrome_options.add_argument(f"--proxy-server=http://{proxy_dict['ip']}:{proxy_dict['port']}")

        if incognito:
            # 无痕
            chrome_options.add_argument("--incognito")

        if headless:
            # 无头
            chrome_options.add_argument("--headless")

        if userdata_file_path:
            chrome_options.add_argument(f"--user-data-dir={userdata_file_path}")
        return chrome_options

    @staticmethod
    def get_undetected_chrome_options(
            headless: bool = True,
            incognito: bool = True,
            no_pop_ups: bool = False,
            page_load_strategy: str = "none",
            no_extensions: bool = False,
            userdata_file_path: str = "",
    ):
        """获取undetected_chrome_options"""
        chrome_options = uc.ChromeOptions()
        chrome_options.add_argument("--disable-gpu")  # 谷歌文档提到需要加上这个属性来规避bug
        chrome_options.add_argument("--hide-scrollbars")  # 隐藏滚动条, 应对一些特殊页面
        chrome_options.add_argument("--disable-infobars")  # 关闭提示
        chrome_options.add_argument("--disable-dev-shm-usage")  # 修复选项卡崩溃
        chrome_options.add_argument("--ignore-certificate-errors")  # 忽略验证：您的连接不是私密连接

        if no_extensions:
            chrome_options.add_argument("--disable-extensions")
            chrome_options.add_experimental_option("useAutomationExtension", False)

        if no_pop_ups:
            # 禁止弹窗加入
            prefs = {
                "profile.default_content_setting_values":
                    {
                        "notifications": 2
                    }
            }
        else:
            prefs = dict()
        prefs["credentials_enable_service"] = False
        prefs["profile.password_manager_enabled"] = False
        chrome_options.add_experimental_option("prefs", prefs)
        chrome_options.set_capability("PageloadStrategy", page_load_strategy)

        if incognito:
            # 无痕
            chrome_options.add_argument("--incognito")

        if headless:
            # 无头
            chrome_options.add_argument("--headless")

        if userdata_file_path:
            chrome_options.add_argument(f"--user-data-dir={userdata_file_path}")
        return chrome_options

    def clear_and_send_keys(self, browser, by: str, value: str, msg):
        """
        点击并发送
        每次都获取是因为有时旧的元素定位不到会报错，所以每次都重新获取元素并完成相应操作
        :param browser: 浏览器对象
        :param by: 通过什么方式定位，如： id，xpath，css_selector
        :param value: 定位的规则
        :param msg: 需要输入的值
        :return:
        """
        by = by.lower()
        self.__get_element(browser, by, value).clear()
        time.sleep(0.1)
        self.__get_element(browser, by, value).send_keys(msg)
        time.sleep(0.3)
        return True

    def find_element(self, browser, by, value):
        """获取元素"""
        return self.__get_element(browser, by, value)

    def find_element_by_xpath(self, browser, value: str):
        """通过xpath获取元素"""
        return self.__get_element(browser, "xpath", value)

    def find_element_by_id(self, browser, value: str):
        """通过id获取元素"""
        return self.__get_element(browser, "id", value)

    def find_elements(self, browser, by, value):
        """获取元素"""
        return self.__get_elements(browser, by, value)

    def find_elements_by_xpath(self, browser, value: str):
        """通过xpath获取元素"""
        return self.__get_elements(browser, "xpath", value)

    def find_elements_by_id(self, browser, value: str):
        """通过id获取元素"""
        return self.__get_elements(browser, "id", value)

    def wait_clickable_element(self, browser, by, value: str, wait=10):
        """等待可点击的元素"""
        if by == 'id':
            return self.wait_clickable_element_by_id(browser, value, wait)
        elif by == 'xpath':
            return self.wait_clickable_element_by_xpath(browser, value, wait)
        return None

    @staticmethod
    def wait_clickable_element_by_xpath(browser, value: str, wait=10):
        """等待可点击的元素"""
        return WebDriverWait(browser, wait, 0.5).until(EC.element_to_be_clickable((By.XPATH, value)))

    @staticmethod
    def wait_clickable_element_by_id(browser, value: str, wait=10):
        """等待可点击的元素"""
        return WebDriverWait(browser, wait, 0.5).until(EC.element_to_be_clickable((By.ID, value)))

    def wait_located_element(self, browser, by, value: str, wait=10):
        """等待元素"""
        if by == 'id':
            return self.wait_located_element_by_id(browser, value, wait)
        elif by == 'xpath':
            return self.wait_located_element_by_xpath(browser, value, wait)
        return None

    @staticmethod
    def wait_located_element_by_xpath(browser, value: str, wait=10):
        """等待元素"""
        return WebDriverWait(browser, wait, 0.5).until(EC.presence_of_element_located((By.XPATH, value)))

    @staticmethod
    def wait_located_element_by_id(browser, value: str, wait=10):
        """等待元素"""
        return WebDriverWait(browser, wait, 0.5).until(EC.presence_of_element_located((By.ID, value)))

    def wait_text_to_be_present(self, browser, by, value: str, text, wait=10):
        """等待元素"""
        if by == 'id':
            return self.wait_text_to_be_present_by_id(browser, value, text, wait)
        elif by == 'xpath':
            return self.wait_text_to_be_present_by_xpath(browser, value, text, wait)
        return None

    @staticmethod
    def wait_text_to_be_present_by_xpath(browser, value: str, text: str, wait=10):
        """等待文字"""
        return WebDriverWait(browser, wait, 0.5).until(EC.text_to_be_present_in_element((By.XPATH, value), text_=text))

    @staticmethod
    def wait_text_to_be_present_by_id(browser, value: str, text: str, wait=10):
        """等待文字"""
        return WebDriverWait(browser, wait, 0.5).until(EC.text_to_be_present_in_element((By.ID, value), text_=text))

    def wait_ele_to_be_selected(self, browser, by, value: str, wait=10):
        """等待选择框"""
        if by == 'id':
            return self.wait_ele_to_be_selected_by_id(browser, value, wait)
        elif by == 'xpath':
            return self.wait_ele_to_be_selected_by_xpath(browser, value, wait)
        return None

    def wait_ele_to_be_selected_by_id(self, browser, value: str, wait=10):
        """等待选择框"""
        return Select(self.wait_clickable_element_by_id(browser, value, wait))

    def wait_ele_to_be_selected_by_xpath(self, browser, value: str, wait=10):
        """等待选择框"""
        return Select(self.wait_clickable_element_by_xpath(browser, value, wait))

    @staticmethod
    def wait_alert_element(browser, wait=10):
        """等待弹窗"""
        return WebDriverWait(browser, wait, 0.5).until(EC.alert_is_present())

    def click_by_js(self, browser, by, value):
        """通过js点击"""
        by = by.lower()
        ele = self.__get_element(browser, by, value)
        browser.execute_script("arguments[0].click();", ele)
        time.sleep(0.5)
        return True

    def scroll_to_ele(self, browser, by, value):
        """滚动到某个元素的位置"""
        ele = self.__get_element(browser, by, value)
        js4 = "arguments[0].scrollIntoView();"
        browser.execute_script(js4, ele)
        return True

    def __get_element(self, browser, by, value):
        """通过不同的方式查找界面元素"""
        element = None
        by = by.lower()
        if by == "id":
            element = browser.find_element(by=By.ID, value=value)
        elif by == "name":
            element = browser.find_element(by=By.NAME, value=value)
        elif by == "xpath":
            element = browser.find_element(by=By.XPATH, value=value)
        elif by == "classname":
            element = browser.find_element(by=By.CLASS_NAME, value=value)
        elif by == "css":
            element = browser.find_element(by=By.CSS_SELECTOR, value=value)
        elif by == "link_text":
            element = browser.find_element(by=By.LINK_TEXT, value=value)
        else:
            self.logger.error("无对应方法，请检查")
        return element

    def __get_elements(self, browser, by, value):
        """通过不同的方式查找所有界面元素"""
        elements = []
        by = by.lower()
        if by == "id":
            elements = browser.find_elements(by=By.ID, value=value)
        elif by == "name":
            elements = browser.find_elements(by=By.NAME, value=value)
        elif by == "xpath":
            elements = browser.find_elements(by=By.XPATH, value=value)
        elif by == "classname":
            elements = browser.find_elements(by=By.CLASS_NAME, value=value)
        elif by == "css":
            elements = browser.find_elements(by=By.CSS_SELECTOR, value=value)
        elif by == "link_text":
            elements = browser.find_elements(by=By.LINK_TEXT, value=value)
        else:
            self.logger.error("无对应方法，请检查")
        return elements

    @staticmethod
    def get_browser_cookies(browser):
        """获取当前浏览器cookie，返回str"""
        cookie_dict = browser.get_cookies()
        cookies = ""
        for cookie in cookie_dict:
            if not cookie["name"].startswith("_"):
                cookies += cookie["name"] + "=" + cookie["value"] + ";"
        return cookies.strip(";")

    def __create_proxyauth_extension(self, proxy_host, proxy_port, proxy_username, proxy_password, scheme="http"):
        """带密码验证的扩展"""
        plugin_path = f"./proxy_auth_plugin_{self.task_name}"

        manifest_json = """
        {
            "version": "1.0.0",
            "manifest_version": 2,
            "name": "Chrome Proxy",
            "permissions": [
                "proxy",
                "tabs",
                "unlimitedStorage",
                "storage",
                "<all_urls>",
                "webRequest",
                "webRequestBlocking"
            ],
            "background": {
                "scripts": ["background.js"]
            },
            "minimum_chrome_version":"22.0.0"
        }
        """

        background_js = string.Template(
            """
            var config = {
                    mode: "fixed_servers",
                    rules: {
                      singleProxy: {
                        scheme: "${scheme}",
                        host: "${host}",
                        port: parseInt(${port})
                      },
                      bypassList: ["foobar.com"]
                    }
                  };

            chrome.proxy.settings.set({value: config, scope: "regular"}, function() {});

            function callbackFn(details) {
                return {
                    authCredentials: {
                        username: "${username}",
                        password: "${password}"
                    }
                };
            }

            chrome.webRequest.onAuthRequired.addListener(
                        callbackFn,
                        {urls: ["<all_urls>"]},
                        ["blocking"]
            );
            """
        ).substitute(
            host=proxy_host,
            port=proxy_port,
            username=proxy_username,
            password=proxy_password,
            scheme=scheme,
        )
        with zipfile.ZipFile(plugin_path, "w") as zp:
            zp.writestr("manifest.json", manifest_json)
            zp.writestr("background.js", background_js)
        return plugin_path

    def close_browser(self, browser):
        """
        关闭浏览器
        """
        try:
            browser.quit()
            time.sleep(0.5)
            if self.__browser_model in ["local", "undetected"] and sys.platform == "linux":
                # 关闭虚拟窗口
                self.__display.stop()
        except Exception as e:
            print(e)
        return True



# -*- coding: utf-8 -*-
"""
操作FTP服务器
目前仅支持下载和删除
"""
import os
import ftplib


def connect_ftp_server(func):
    def wrapper(self, *args, **kwargs):
        try:
            self.ftp.cwd("/")
        except:
            self.ftp = self.ftp_connect()
        return func(self, *args, **kwargs)

    return wrapper


class FtpTool(object):

    def __init__(self, ip, port, username, password):
        self.ftp_server = ip  # ftp主机IP
        self.port = port  # ftp端口
        self.username = username  # 登陆用户名
        self.password = password  # 登陆密码
        self.ftp = self.ftp_connect()

    def ftp_connect(self):
        """ftp连接"""
        ftp = ftplib.FTP()
        try:
            ftp.connect(self.ftp_server, self.port)
            ftp.login(self.username, self.password)
        except:
            raise IOError('FTP login failed!!!')
        else:
            print(ftp.getwelcome())
            return ftp

    @connect_ftp_server
    def upload_file(self, remote_path, local_path):
        """
        上传文件
        """
        self.ftp.cwd(remote_path)
        buf_size = 1024 * 1024
        with open(local_path, "rb") as f:
            self.ftp.storbinary('STOR %s' % os.path.basename(local_path), f, buf_size)  # 上传文件到ftp服务器
        return True

    @connect_ftp_server
    def download_file(self, remote_path, local_path):
        """单个文件下载到本地"""
        with open(local_path, 'wb') as file_handle:
            self.ftp.retrbinary(f"RETR {remote_path}", file_handle.write)
            print(f"下载至: {local_path}")
        return True

    @connect_ftp_server
    def download_files(self, remote_path, local_path):
        """下载整个目录下的文件,包括子目录文件"""
        if not os.path.exists(local_path):
            os.makedirs(local_path)
        self.ftp.cwd(remote_path)

        for i, file_name in enumerate(self.ftp.nlst()):
            if not file_name.endswith(".zip"):
                continue
            ftp_path = os.path.join(remote_path, file_name)
            local = os.path.join(local_path, file_name)

            if self.__is_filedir(file_name=file_name):  # 判断是否为子目录
                if not os.path.exists(local):
                    os.makedirs(local)
                self.download_files(remote_path=ftp_path, local_path=local)
            else:
                self.download_file(remote_path=ftp_path, local_path=local)
        self.ftp.cwd('..')
        return True

    @connect_ftp_server
    def move_file(self, from_path, to_path):
        """移动文件"""
        self.ftp.rename(from_path, to_path)
        return True

    @connect_ftp_server
    def delete_file(self, remote_file_path):
        """删除文件夹中的所有文件"""
        try:
            self.ftp.delete(remote_file_path)
            print(f"{remote_file_path} 删除成功")
        except Exception as e:
            print(e)

    @connect_ftp_server
    def delete_files(self, remote_path):
        """清空文件夹"""
        self.ftp.cwd(remote_path)
        for file_name in self.ftp.nlst():
            if not file_name.endswith(".zip"):
                continue
            if self.__is_filedir(file_name=file_name):  # 判断是否为子目录
                self.delete_files(file_name)
            else:
                self.ftp.delete(file_name)

    @connect_ftp_server
    def ftp_disconnect(self):
        """退出FTP连接"""
        self.ftp.quit()

    def __is_filedir(self, file_name):
        """
        判断是否是文件夹
        :param file_name: 文件名/文件夹名
        :return:返回字符串“File”为文件，“Dir”问文件夹，“Unknow”为无法识别
        """
        rec = ""
        try:
            rec = self.ftp.cwd(file_name)  # 需要判断的元素
            self.ftp.cwd("..")  # 如果能通过路劲打开必为文件夹，在此返回上一级
        except Exception as e:
            rec = e  # 不能通过路劲打开必为文件，抓取其错误信息
        finally:
            if "550 Failed to change directory" in str(rec):
                return False
            elif "Directory successfully changed." in str(rec):
                return True
            else:
                return False



# -*- coding: utf-8 -*-
"""
SetQueue: 去重队列，继承自queue.Queue，改写 _init 中 deque 为 set
OrderedSetQueue: 有序去重队列，继承自queue.Queue, 改写 _init 中 deque 为 OrderedSet
"""
from queue import Queue
from orderedset import OrderedSet


class SetQueue(Queue):
    """
    去重队列
    """

    def _init(self, maxsize):
        # 替换父类 deque 为 set
        self.queue = set()

    def _put(self, item):
        self.queue.add(item)

    def _get(self):
        return self.queue.pop()


class OrderedSetQueue(Queue):
    """
    有序去重队列
    """

    def _init(self, maxsize):
        self.queue = OrderedSet()

    def _put(self, item):
        self.queue.add(item)

    def _get(self):
        return self.queue.pop()



# -*- coding: utf-8 -*-
"""
基于selenium的爬虫类
"""
import urllib3

from wise_utils.common import get_logger

urllib3.disable_warnings()


class BaseSpider:
    def __init__(self, task_name, log_file_path=None, log_level="DEBUG"):
        self.task_name = task_name.lower()
        self.logger = get_logger(name=task_name, path=log_file_path, log_level=log_level)

    def get_proxy_ip(self):
        return {"http": "http://", "https": "http://"}



# -*- coding: utf-8 -*-
__all__ = ["AllException", "FtpTool", "get_logger", "UtilMail", "retry", "SetQueue", "OrderedSetQueue", "inline"]

from wise_utils.common.exceptions import AllException
from wise_utils.common.ftp import FtpTool
from wise_utils.common.log import get_logger
from wise_utils.common.mail import UtilMail
from wise_utils.common.retry_decorator import retry
from wise_utils.common.setqueue import SetQueue, OrderedSetQueue
from yagmail import inline



# -*- coding: utf-8 -*-
import redis


def get_redis_client(redis_config):
    db_redis = redis.StrictRedis(connection_pool=redis.ConnectionPool(**redis_config, decode_responses=True)) if redis_config else None
    return db_redis



# -*- coding: utf-8 -*-
"""
重试机制，此装饰器只适用于装饰类中的方法，并且，调用的也是类中的方法，关键处在于：是否接收并传递self
"""
import time

from inspect import isfunction


def __after_exception(*args, **kwargs):
    pass


def __retry_error_callback(*args, **kwargs):
    pass


def retry(max_retry=3, current_times=0, wait=0, after_exception=__after_exception, retry_error_callback=__retry_error_callback, remark=""):
    """
    装饰器实现重试机制
    :param max_retry: 重试次数：默认3次，-1代表一直重试
    :param current_times: 当前重试次数
    :param wait: 每次重试的时间间隔
    :param after_exception: 重试睡眠前调用的函数
    :param retry_error_callback: 重试结束仍然失败时的回调函数
    :param remark: 备注
    :return:
    """

    def catch_exception(func):
        def wrapper(self, *args, **kwargs):
            nonlocal max_retry, current_times
            try:
                # 调用func
                res = func(self, *args, **kwargs)
                current_times = 0
                return res
            except Exception as e:
                current_times += 1
                # 睡眠前调用
                if isfunction(after_exception):
                    after_exception(self, *args, **kwargs, remark=remark)

                # 重试间隔
                time.sleep(wait)

                if max_retry == -1:
                    return wrapper(self, *args, **kwargs)
                elif current_times > max_retry:
                    if isfunction(retry_error_callback):
                        # 重试结束，执行回调函数
                        return retry_error_callback(self, *args, **kwargs, remark=remark)
                    else:
                        # 没有指定回调函数时，弹出错误
                        raise e
                else:
                    current_times += 1
                    return wrapper(self, *args, **kwargs)

        # 设置current_times参数值
        def set_current_times(value):
            nonlocal current_times
            current_times = value

        wrapper.set_current_times = set_current_times
        return wrapper

    return catch_exception



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

VERSION = "0.3.6"

setuptools.setup(
    name="wise-utils",
    version=VERSION,
    author="wise-python",
    author_email="duke.du@uifox.com.cn",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitee.com/duquan1995/wise-utils.git",
    project_urls={
        "Bug Tracker": "https://github.com/pypa/sampleproject/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    python_requires=">3",
    install_requires=[
        "yagmail>=0.15.277",
        "orderedset>=2.0.3",
        "requests>=2.25.1",
        "urllib3>=1.26.7",
        "selenium>=4.1.0",
        "pyppeteer~=1.0.2",
        "redis",
        "PyMySQL>=1.0.2",
        "nacos-sdk-python>=0.1.6",
        "better-exceptions>=0.3.3",
        "loguru>=0.6.0",
        "PyVirtualDisplay>=2.2",
        "undetected_chromedriver~=3.1.5",
        "webdriver_manager~=3.5.4"
    ]
)



# -*- coding: utf8 -*-
import time
import pymysql
import traceback

from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from dbutils.persistent_db import PersistentDB
from dbutils.pooled_db import PooledDB


class StoreMysqlPool(object):
    """
    mysql读写相关操作，pymysql，使用PooledDB连接池
    Args:
        host:数据库ip
        user:数据库用户名
        password:数据库用户密码
        db:数据库名
        port:数据库端口，默认3306
        mincached:最少的空闲连接数， 默认为1
        charset:数据库编码，默认utf8
    """

    def __init__(self, host="", user="", password="", db="", port=3306, mincached=1, pattern=1, charset="utf8"):
        self.host = host

        if pattern == 1:
            self.pool = PooledDB(pymysql, mincached, host=host, user=user, passwd=password, db=db, port=port, charset=charset)
            self.close_able = True
        elif pattern == 2:
            self.pool = PersistentDB(pymysql, host=host, user=user, passwd=password, db=db, port=port, charset=charset)
            self.close_able = False
        self._last_use_time = time.time()

    def close(self, db):
        if db is not None:
            db.close()

    def connect(self):
        return self.pool.connection()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_all()

    @staticmethod
    def _cursor(db):
        # return db.cursor(pymysql.cursors.DictCursor)
        return db.cursor()

    def query(self, sql):
        """
        根据sql查询
        Returns： 数组的数组，外层数组元素为一行，内存数组元素为一行的一列
        """
        db = self.connect()
        cur = self._cursor(db)
        rows = []
        try:
            cur.execute(sql)
            db.commit()
            rows = cur.fetchall()
        except pymysql.OperationalError:
            print(traceback.format_exc())
        except Exception:
            print(traceback.format_exc())
        finally:
            cur.close()
            self.close(db)
        return rows

    def count(self, tb):
        """
        返回某表的行数
        Args:
            tb:字符串，表名称
        """
        sql = 'select count(1) from %s' % tb
        results = self.query(sql)
        if len(results) == 1 and len(results[0]) == 1:
            return int(results[0][0])

    def do(self, sql, flag='lastrowid'):
        """
        执行sql，insert/delete/update操作
        Args:
            sql:要执行的sql
            flag:返回值类型，flag=lastrowid返回lastrowid，flag=rowcount返回rowcount
        """
        db = self.connect()
        cur = self._cursor(db)
        r = None
        try:
            cur.execute(sql)
            db.commit()
            r = 1
            if flag == 'lastrowid':
                r = cur.lastrowid
            elif flag == 'rowcount':
                r = cur.rowcount
        except pymysql.OperationalError:
            print("time: %s; Error on %s: %s" % (str(datetime.now()), self.host, sql[0: 200]))
            print(traceback.format_exc())
            r = -1
        except Exception:
            print("time: %s; Error to MySQL on %s: %s" % (str(datetime.now()), self.host, sql[0: 200]))
            print(traceback.format_exc())
            r = -1
        finally:
            cur.close()
            self.close(db)
        return r

    def save(self, table, data, mapping_fields=dict()):
        """
        将字典直接insert到数据库
        Args:
            table:字符串，插入目标表的名称
            data:字典格式，key为字段名称，value为字段值，如{'id':'1','name':'temp'}
            mapping_fields: 用于保持data字典的key与数据库字段的对应关系，
                            如果结果字典的某个key不包含在mapping_fields中，则将直接使用key作为字段名
        """

        if len(data) <= 0:
            return -1
        try:
            fields = ''
            values = ''
            for d in data:
                if d in mapping_fields:
                    fields += "`%s`," % (str(mapping_fields[d]))
                else:
                    fields += "`%s`," % (str(d))
                values += "'%s'," % (pymysql.converters.escape_string(str(data[d])))
            if len(fields) <= 0 or len(values) <= 0:
                return -1
            sql = "insert ignore into %s(%s) values(%s)" % (table, fields[:-1], values[:-1])
            return self.do(sql)
        except Exception:
            print(traceback.format_exc())
            return -1

    def save_many(self, sql, values, flag='lastrowid'):
        """
        执行sql，insert/delete/update操作
        :param table: 'insert into table(id,name) values(%s,%s)'
        :param names: 字符串
        :param values:[{1,2},{1,2},...]  是一个列表，列表中的每一个元素必须是元组！！！
        :param flag:
        :return:
        """
        db = self.connect()
        cur = self._cursor(db)
        r = None
        try:
            cur.executemany(sql, values)
            db.commit()
            r = 1
            if flag == 'lastrowid':
                r = cur.lastrowid
            elif flag == 'rowcount':
                r = cur.rowcount

        except pymysql.OperationalError:
            print("Error connecting to MySQL on %s: %s" % (self.host, sql[0:255]))
            print(traceback.format_exc())
            r = -1
        except Exception:
            print("Error connecting to MySQL on %s: %s" % (self.host, sql[0:255]))
            print(traceback.format_exc())
            r = -1
        finally:
            cur.close()
            self.close(db)
        return r

    def update(self, table, data, field, mapping_fields=dict()):
        """
        将字典直接update到数据库
        Args:
            table:字符串，更新目标表的名称
            data:字典格式，key为字段名称，value为字段值，如{'id':'1','name':'temp'}
            field:唯一索引字段，即根据该字段判断是否为同一条记录，作为where条件
            mapping_fields: 用于保持data字典的key与数据库字段的对应关系，
                            如果结果字典的某个key不包含在mapping_fields中，则将直接使用key作为字段名
        """
        if len(data) <= 0:
            return -1
        else:
            try:
                values = ''
                field_value = None
                for d in data:
                    key = d
                    if d in mapping_fields:
                        key = mapping_fields[d]
                    if key == field:
                        field_value = data[d]
                    values += "%s='%s'," % (str(key), pymysql.converters.escape_string(str(data[d])))
                if len(values) <= 0 or field_value is None:
                    return -1
                sql = "update " + table + " set " + values[:-1] + " where " + field + "='" + pymysql.converters.escape_string(str(field_value)) + "'"
                return self.do(sql, flag='rowcount')
            except Exception:
                print(traceback.format_exc())
                return -1

    def saveorupdate(self, table, data, field, mapping_fields=dict()):
        """
        将字典更新到数据库，如果已存在则update，不存在则insert
        Args:
            table:字符串，更新目标表的名称
            data:字典格式，key为字段名称，value为字段值，如{'id':'1','name':'temp'}
            field:唯一索引字段，即根据词字段判断是否为同一条记录，作为where条件
            mapping_fields: 用于保持data字典的key与数据库字段的对应关系，
                            如果结果字典的某个key不包含在mapping_fields中，则将直接使用key作为字段名
        """
        if len(data) <= 0:
            return -1
        try:
            field_value = None
            if field in data:
                field_value = data[field]
            else:
                for key in mapping_fields:
                    if mapping_fields[key] == field and key in data:
                        field_value = data[key]
            if field_value is None:
                return -1
            querysql = "select count(1) from " + table + " where " + field + "='" + pymysql.converters.escape_string(str(field_value)) + "'"
            ed = self.query(querysql)
            if ed and ed[0][0] > 0:
                return self.update(table, data, field, mapping_fields)
            else:
                return self.save(table, data, mapping_fields)
        except Exception:
            print(traceback.format_exc())
            return -1

    def create_table(self, tb_name, likeTable):
        """
        创建数据表格
        :param tb_name: table's name
        :return: 1 ok , -1 error
        """
        # sql = "SELECT table_name FROM information_schema.TABLES WHERE table_name ='{}';".format(tb_name)
        sql = "show tables;"
        datas = self.query(sql)
        if tb_name in [str(i[0]) for i in datas]:
            # 存在
            sql = "select * from {} limit 10;".format(tb_name)
            if not self.query(sql):
                print("{} 是空表, 正在truncate.....".format(tb_name))
                sql = " truncate {};".format(tb_name)
                datas = self.query(sql)
            else:
                print("表格存在：{}".format(tb_name))

        else:
            # 不存在 创建 清空
            print("{} 表格不存在，正在创建....".format(tb_name))
            sql = "create table {} like {};".format(tb_name, likeTable)
            datas = self.query(sql)
            sql = " truncate {};".format(tb_name)
            datas = self.query(sql)
        return 1

    def dicts_2_OrderDicts(self, list_dicts):
        alist = []
        # list_dicts = [list_dicts] if isinstance(list_dicts, dict) else list_dicts
        ks = list_dicts[0].keys()  # 有序
        order_dict = OrderedDict()
        for adict in list_dicts:
            for k in ks:
                order_dict[k] = adict[k]
            alist.append(deepcopy(order_dict))
        return alist

    def insert_many(self, table, list_dicts, conflict=''):
        '''
        [批量插入]
        :param store_name:
        :param table:
        :param list_dicts:
        :param conflict: 冲突更新字段 local_date=CURRENT_DATE() or ['local_date','fields']
        :return:
        '''
        try:
            if len(list_dicts):
                orderDicts = self.dicts_2_OrderDicts(list_dicts)
                ks = orderDicts[0].keys()  # 获取键值
                if conflict and isinstance(conflict, str):
                    insert_sql = "insert ignore into {} ({}) value ({}) ON DUPLICATE KEY UPDATE {}".format(table, ", ".join(ks), ("%s," * len(ks)).strip(","), conflict)
                else:
                    insert_sql = "insert ignore into {} ({}) values ({})".format(table, ", ".join(ks), ("%s," * len(ks)).strip(","))
                values = []
                for adict in list_dicts:
                    values.append(tuple(adict.values()))
                res = self.save_many(insert_sql, values, flag="rowcount")
                return res
            else:
                return -3
        except:
            print(traceback.format_exc())
            return -2

    def close_all(self):
        """
        关闭连接池全部连接
        请在确保连接不再需要时确认
        :return:
        """
        if self.close_able:
            self.pool.close()


def get_mysql_client(host="", user="", password="", db="", port=3306, mincached=1, pattern=1, charset="utf8"):
    return StoreMysqlPool(host=host, user=user, password=password, db=db, port=port, mincached=mincached, pattern=pattern, charset=charset)
